@@ -1,3 +1,45 @@
+## This file is part of mvSLOUCH
+
+## This software comes AS IS in the hope that it will be useful WITHOUT ANY WARRANTY, 
+## NOT even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
+## Please understand that there may still be bugs and errors. Use it at your own risk. 
+## We take no responsibility for any errors or omissions in this package or for any misfortune 
+## that may befall you or others as a result of its use. Please send comments and report 
+## bugs to Krzysztof Bartoszek at krzbar@protonmail.ch .
+
+.calc_species_mean<-function(evolmodel,expmtA,exptjA,regimes,vX0=NULL,vY0=NULL,mPsi=NULL,A1B=NULL,mPsi0=NULL){
+    vMean=switch(evolmodel,
+	bm=vX0, ## BM with drift here also
+	ouch=.calc.mean.ouch.mv(expmtA,vY0,mPsi,mPsi0,exptjA,regimes),
+#	slouch=(),
+	mvslouch=.calc.mean.slouch.mv(expmtA,A1B,vY0,vX0,mPsi,mPsi0,exptjA,regimes)
+    )
+    vMean
+}
+
+.calc_species_covariance<-function(evolmodel,species_time,lAcalcs,lScalcs,method="minus.v",tol=1e-10,stationary=FALSE,b_correct_forPD=TRUE){
+    mCov=switch(evolmodel, 
+	bm=species_time*lScalcs$S22, ## BM with drift here also
+	ouch=.calc.cov.ouch.mv(species_time,lAcalcs,lScalcs,stationary=stationary,method=method,b_correct_forPD=b_correct_forPD),
+#	slouch=(),
+	mvslouch=.calc.cov.slouch.mv(species_time,lAcalcs,lScalcs,tol=tol,method=method,b_correct_forPD=b_correct_forPD)
+    )
+    if (b_correct_forPD&&((!matrixcalc::is.symmetric.matrix(mCov)) || (!matrixcalc::is.positive.definite(mCov)))){
+	tryCatch({mCov<-as.matrix(Matrix::nearPD(mCov)$mat)},error=function(e){.my_message(e,FALSE)})    
+    }
+    mCov
+}
+
+.calc.mean.bm.mv<-function(vX0){vX0}
+.calc.cov.bm.mv<-function(species_time,StS,b_correct_forPD=TRUE){
+    mCov<-species_time*StS
+    if (b_correct_forPD && ((!matrixcalc::is.symmetric.matrix(mCov)) || (!matrixcalc::is.positive.definite(mCov)))){
+        tryCatch({mCov<-as.matrix(Matrix::nearPD(mCov)$mat)},error=function(e){.my_message(e,FALSE)})
+    }
+    mCov
+}
+    
+
 .calc.mean.ouch.mv<-function(expmtA,vY0,mPsi,mPsi0,exptjA=NULL,regimes=NULL){
 ## t is not needed to be passed, it is in the expmtA not explicitely
 ## vPsia has to be given as a vector
@@ -8,14 +50,13 @@
 	vPsiSum<-apply(matrix(sapply(1:(length(exptjA)-1),function(x){(exptjA[[x+1]]-exptjA[[x]])%*%mPsi[,regimes[x]]},simplify=TRUE),nrow=length(vY0)),1,sum) 
 	vMean<-vMean+vPsiSum ## a trick here to reuse the calculations from vMean
     }else{vMean<-vMean+(diag(1,kY,kY)-expmtA)%*%mPsi[,1]} ## there is only one optimum ie one Psi
-#    vMean<-expmtA%*%vMean
     if (!(is.null(mPsi0))){vMean<-vMean+(diag(1,kY,kY)-expmtA)%*%mPsi0}
     vMean[which(abs(vMean)<1e-15)]<-0
     names(vMean)<-NULL
     vMean
 }
 
-.calc.cov.ouch.mv<-function(t,lAcalcs,lScalcs,stationary=FALSE,method="minus.v"){
+.calc.cov.ouch.mv<-function(t,lAcalcs,lScalcs,stationary=FALSE,method="minus.v",b_correct_forPD=TRUE){
     invA<-lAcalcs$invA
     kY<-nrow(lAcalcs$A)
     mCov<-matrix(NA,kY,kY)
@@ -35,6 +76,9 @@
     mCov[which(abs(mCov)<1e-15)]<-0
     colnames(mCov)<-NULL
     rownames(mCov)<-NULL
+    if (b_correct_forPD && ((!matrixcalc::is.symmetric.matrix(mCov)) || (!matrixcalc::is.positive.definite(mCov)))){
+        tryCatch({mCov<-as.matrix(Matrix::nearPD(mCov)$mat)},error=function(e){.my_message(e,FALSE)})
+    }
     mCov
 }
 
@@ -46,7 +90,6 @@
     if (!(is.null(exptjA))){
 	## we have for length - 1 as we only take differences
 	vPsiSum<-apply(matrix(sapply(1:(length(exptjA)-1),function(x){(exptjA[[x+1]]-exptjA[[x]])%*%mPsi[,regimes[x]]},simplify=TRUE),nrow=length(vY0)),1,sum) 
-#	vMean[1:kY,]<-vMean[1:kY,]+expmtA%*%(vY0+vPsiSum-vMean[1:kY,])
 	vMean[1:kY,]<-vMean[1:kY,]+expmtA%*%(vY0-vMean[1:kY,])+vPsiSum
     }else{vMean[1:kY,]<-vMean[1:kY,]+expmtA%*%(vY0-mPsi[,1]-vMean[1:kY,])+mPsi[,1]} ## there is only one optimum ie one Psi
     if (!(is.null(mPsi0))){vMean[1:kY,]<-vMean[1:kY,]+(diag(1,kY,kY)-expmtA)%*%mPsi0}
@@ -56,7 +99,7 @@
 }
 
 
-.calc.cov.slouch.mv<-function(t,lAcalcs,lScalcs,tol=1e-10,method="minus.v"){
+.calc.cov.slouch.mv<-function(t,lAcalcs,lScalcs,tol=1e-10,method="minus.v",b_correct_forPD=TRUE){
 ## once again invP can be calculated from eigA but more effective to do this once
 ## same with invA
 ## the idea of the function is just to glue everything together and only calculate time depedent bits
@@ -64,7 +107,7 @@
     invA<-lAcalcs$invA
     kY<-nrow(A1B)
     kX<-ncol(A1B)
-    tA1B<-t(A1B) ## slightly more effective no need to call t()
+    tA1B<-t(A1B) 
     exptA<-.calc.exptA(t,lAcalcs)
     expmtA<-.calc.exptA(-t,lAcalcs)
     A1BS22tA1B<-A1B%*%lScalcs$S22%*%tA1B
@@ -99,6 +142,10 @@
     mCov<-(mCov+t(mCov))/2            
     colnames(mCov)<-NULL
     rownames(mCov)<-NULL
+    if (b_correct_forPD&&((!matrixcalc::is.symmetric.matrix(mCov)) || (!matrixcalc::is.positive.definite(mCov)))){
+        tryCatch({mCov<-as.matrix(Matrix::nearPD(mCov)$mat)},error=function(e){.my_message(e,FALSE)})
+    }
+
     mCov
 }
 
